;; Generated package description from emms.el  -*- no-byte-compile: t -*-
(define-package "emms" "25.0.20260113.142250" "The Emacs Multimedia System" '((cl-lib "0.5") (nadvice "0.3") (seq "0")) :commit "1d48a1133db2670d839f2cf2aff17d6c32aeb15f" :authors '(("Jorgen Schäfer" . "forcer@forcix.cx")) :maintainer '("Yoni Rabkin" . "yrk@gnu.org") :keywords '("emms" "mp3" "ogg" "flac" "music" "mpeg" "video" "multimedia") :url "https://www.gnu.org/software/emms/")
